﻿using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using static AugmentedEquipmentItems;
using HarmonyLib;
using Steamworks;
using UnityEngine;

public class AugmentedItems : Mod {
  private Harmony harmony;
  private string instanceId = "com.bahamut.augmenteditems";
  private static AssetBundle asset;

  private static List<Item_Base> addedItems = new List<Item_Base>();
  private static Animator cachedLocalAnimator = null;
  private static Dictionary<string, ItemConnection> addedConnectionDictionary = new Dictionary<string, ItemConnection>();
  private static List<AnimationConnection> addedAnimationConnectionList = new List<AnimationConnection>();
  private static Dictionary<string, GameObject> rightHandGameObjects = new Dictionary<string, GameObject> {
    ["hook"] = null
  };
  private static Dictionary<string, AnimationConnection> animationConnectionTemplates = new Dictionary<string, AnimationConnection> {
    ["hookThrow"] = null,
    ["hookGather"] = null
  };

  public IEnumerator Start() {
    AssetBundleCreateRequest request = AssetBundle.LoadFromMemoryAsync(GetEmbeddedFileBytes("com.bahamut.augmenteditems.assets"));
    yield return request;
    asset = request.assetBundle;

    LoadAugmentedItems();

    harmony = new Harmony(instanceId);
    harmony.PatchAll(Assembly.GetExecutingAssembly());
    Debug.Log("Mod AugmentedItems has been loaded!");
  }

  public void OnModUnload() {
    harmony.UnpatchAll();
    UnloadItems();
    UnloadAllInstantiatedObjects();
    asset.Unload(true);
    Debug.Log("Mod AugmentedItems has been unloaded!");
  }

  // Helper Methods
  private static void LoadRightHandComponents(Network_Player networkPlayer) {
    Transform rightHandTransform = networkPlayer.rightHandParent;
    foreach (Transform rightHandItemTransform in rightHandTransform) {
      GameObject rightHandGameObject = rightHandItemTransform.gameObject;
      if (rightHandGameObjects["hook"] == null) {
        Component hookScript = rightHandGameObject.GetComponent(typeof(Hook));
        if (hookScript != null) {
          rightHandGameObjects["hook"] = Instantiate(rightHandGameObject);
          rightHandGameObjects["hook"].SetActive(false);
        }
      }
    }
  }

  private static void CreateAnimationConnectionTemplates(Network_Player networkPlayer, List<AnimationConnection> animationConnections) {
    AnimationConnection connectionTemplate = null;
    foreach (AnimationConnection connection in animationConnections) {
      if (connection.name.Equals("Hook_Throw | Animation_HookThrow")) {
        connectionTemplate = new AnimationConnection();
        CopyAnimationConnection(connectionTemplate, connection);
        animationConnectionTemplates["hookThrow"] = connectionTemplate;
      }
      if (connection.name.Equals("Gather | OnHookGather")) {
        connectionTemplate = new AnimationConnection();
        CopyAnimationConnection(connectionTemplate, connection);
        animationConnectionTemplates["hookGather"] = connectionTemplate;
      }
    }
  }

  private static void CopyAnimationConnection(AnimationConnection connectionTemplate, AnimationConnection targetConnection) {
    connectionTemplate.name = targetConnection.name;
    connectionTemplate.layerName = targetConnection.layerName;
    connectionTemplate.stateName = targetConnection.stateName;
    connectionTemplate.functionName = targetConnection.functionName;
    connectionTemplate.functionParameter = targetConnection.functionParameter;
    connectionTemplate.timeOfEvent = targetConnection.timeOfEvent;
    connectionTemplate.active = targetConnection.active;
    connectionTemplate.calledEvent = targetConnection.calledEvent;
    connectionTemplate.lastNormalizedTime = targetConnection.lastNormalizedTime;
  }

  private static void WireUpComponents() {

  }

  private void LoadAugmentedItems() {
    foreach(KeyValuePair<string, ItemTemplate> kvp in AugmentedEquipmentItems.AllAugmentedEquipmentItems) {
      ItemTemplate itemTemplate = kvp.Value;
      if (itemTemplate.uniqueName != null) {
        CreateEquipment(
          itemTemplate.baseItemReference,
          itemTemplate.uniqueName,
          itemTemplate.displayName,
          itemTemplate.uniqueIndex,
          itemTemplate.description,
          itemTemplate.maxUses,
          itemTemplate.equipSlotType,
          itemTemplate.craftingCategory,
          itemTemplate.subCategory,
          itemTemplate.subCategoryOrder,
          itemTemplate.recipe,
          itemTemplate.spriteName
        );
      }
    }
  }

  private void CreateEquipment(string baseItemName, string uniqueName, string displayName, int uniqueIndex, string description, int maxUses, EquipSlotType slotType, CraftingCategory craftingCategory, string subCategory, int subCategoryOrder, Dictionary<string, int> recipe, string spriteName = null) {
    string spriteAsset = (spriteName != null) ? spriteName : uniqueName;
    Sprite augmentedItemSprite = asset.LoadAsset<Sprite>(spriteAsset);

    Item_Base baseEquipmentItem = ItemManager.GetItemByName(baseItemName);
    Item_Base newEquipmentItem = UnityEngine.Object.Instantiate(baseEquipmentItem);

    // Update the properties the item -- overriding the initialized property so we can set maxUses.
    if (baseItemName.Equals("Backpack")) {
      Traverse.Create(newEquipmentItem).Property("UniqueName").SetValue(uniqueName);
      Traverse.Create(newEquipmentItem).Property("UniqueIndex").SetValue(uniqueIndex);
      ItemInstance_Inventory itemInstanceInventory = new ItemInstance_Inventory(null, "", 1);
      Traverse.Create(newEquipmentItem).Field("settings_Inventory").SetValue(itemInstanceInventory);
    } else {
      Traverse.Create(newEquipmentItem).Field("hasBeenInitialized").SetValue(false);
      newEquipmentItem.Initialize(uniqueIndex, uniqueName, maxUses);
    }
    Traverse.Create(newEquipmentItem).Field("settings_Inventory").Property("DisplayName").SetValue(displayName);
    Traverse.Create(newEquipmentItem).Field("settings_Inventory").Property("Description").SetValue(description);
    Traverse.Create(newEquipmentItem).Field("settings_Inventory").Property("Sprite").SetValue(augmentedItemSprite);
    ItemInstance_Equipment equipmentSlotType = new ItemInstance_Equipment(slotType);
    Traverse.Create(newEquipmentItem).Field("settings_equipment").SetValue(equipmentSlotType);

    ItemInstance_Recipe itemInstanceRecipe = new ItemInstance_Recipe(craftingCategory, false, false, subCategory, subCategoryOrder);
    newEquipmentItem.settings_recipe = itemInstanceRecipe;

    // Create the recipe for the item.
    List<CostMultiple> itemRecipe = new List<CostMultiple>();
    foreach (KeyValuePair<string, int> kvp in recipe) {
      Item_Base item = ItemManager.GetItemByName(kvp.Key);
      CostMultiple itemRecipeQuantity = new CostMultiple(new Item_Base[] { item }, kvp.Value);
      itemRecipe.Add(itemRecipeQuantity);
    }
    newEquipmentItem.settings_recipe.NewCost = itemRecipe.ToArray();
    addedItems.Add(newEquipmentItem);

    Debug.Log($"Successfully Added {newEquipmentItem.UniqueName}");
    RAPI.RegisterItem(newEquipmentItem);
  }

  private void UnloadItems() {
    List<Item_Base> allItems = ItemManager.GetAllItems();
    foreach (Item_Base item in addedItems) {
      Debug.Log($"Unloading Item {item.UniqueName}...");
      allItems.Remove(item);
    }
    addedItems.Clear();
  }

  private static void UnloadAllInstantiatedObjects() {
    List<string> rightHandGameObjectsKeys = new List<string>(rightHandGameObjects.Keys);
    foreach (string objectKey in rightHandGameObjectsKeys) {
      if (rightHandGameObjects[objectKey] != null) {
        Destroy(rightHandGameObjects[objectKey]);
        rightHandGameObjects[objectKey] = null;
      }
    }
  }

  private static void UnloadAllAddedConnections(Network_Player networkPlayer) {
    Dictionary<string, ItemConnection> connectionDictionary = Traverse.Create(networkPlayer.PlayerItemManager.useItemController).Field("connectionDictionary").GetValue() as Dictionary<string, ItemConnection>;
    GameObject playerAnimatorGameObject = networkPlayer.Animator.gameObject;
    AnimationEventCaller animationEventCaller = playerAnimatorGameObject.GetComponent<AnimationEventCaller>();
    List<AnimationConnection> animationConnections = Traverse.Create(animationEventCaller).Field("connections").GetValue() as List<AnimationConnection>;
    
    // Remove the MonoBehaviour_Network from the NetworkUpdateManager as well before closing out as a local player as well.
    foreach (KeyValuePair<string, ItemConnection> kvp in addedConnectionDictionary) {
      GameObject connectionItemGameObject = kvp.Value.obj;
      MonoBehaviour_ID_Network itemIdNetwork = connectionItemGameObject.GetComponent<MonoBehaviour_ID_Network>();
      if (itemIdNetwork is MonoBehaviour_Network) {
        MonoBehaviour_Network itemNetwork = itemIdNetwork as MonoBehaviour_Network;
        NetworkUpdateManager.RemoveBehaviour(itemNetwork);
      } else {
        if (itemIdNetwork != null) {
          NetworkIDManager.RemoveNetworkID(itemIdNetwork);
        }
      }
    }

    foreach (KeyValuePair<string, ItemConnection> kvp in addedConnectionDictionary) {
      connectionDictionary.Remove(kvp.Key);
    }

    foreach (AnimationConnection animationConnection in addedAnimationConnectionList) {
      animationConnections.Remove(animationConnection);
    }

    addedConnectionDictionary.Clear();
    addedAnimationConnectionList.Clear();
  }

  // Patches
  [HarmonyPatch(typeof(Network_Player), "OnPlayerCreated")]
  class Network_PlayerOnPlayerCreated_Patch {
    private static void Postfix(Network_Player __instance) {
      Network_Player networkPlayer = __instance;
      Dictionary<string, ItemConnection> connectionDictionary = Traverse.Create(networkPlayer.PlayerItemManager.useItemController).Field("connectionDictionary").GetValue() as Dictionary<string, ItemConnection>;

      GameObject playerAnimatorGameObject = networkPlayer.Animator.gameObject;
      AnimationEventCaller animationEventCaller = playerAnimatorGameObject.GetComponent<AnimationEventCaller>();
      List<AnimationConnection> animationConnections = Traverse.Create(animationEventCaller).Field("connections").GetValue() as List<AnimationConnection>;

      Transform playerRightHandTransform = networkPlayer.rightHandParent;

      foreach (Item_Base itemBase in addedItems) {
        if (itemBase.UniqueName.Contains("Hook")) {
          // This excerpt is for adding in Custom Hooks -- as a result, this will iterate through the existing Player Component to find a hook template.
          GameObject hookTemplateGameObject = null;
          foreach (Transform rightHandItem in playerRightHandTransform) {
            Hook templateHookScript = rightHandItem.gameObject.GetComponent<Hook>();
            if (templateHookScript != null) {
              hookTemplateGameObject = rightHandItem.gameObject;
              // Debug.Log("-- Found a Template Hook Early -- Breaking Out...");
              break;
            }
          }

          // This excerpt is for grabing the animation connections and creating them for the new hook we're adding.
          AnimationConnection hookThrowAnimation = null;
          AnimationConnection hookGatherAnimation = null;
          foreach (AnimationConnection connection in animationConnections) {
            if (connection.name.Equals("Hook_Throw | Animation_HookThrow") && hookThrowAnimation == null) {
              hookThrowAnimation = new AnimationConnection();
              CopyAnimationConnection(hookThrowAnimation, connection);
              // Debug.Log("-- Finished creating the Hook_Throw | Animation_HookThrow.");
            }
            if (connection.name.Equals("Gather | OnHookGather") && hookGatherAnimation == null) {
              hookGatherAnimation = new AnimationConnection();
              CopyAnimationConnection(hookGatherAnimation, connection);
              // Debug.Log("-- Finished creating the Gather | OnHookGather.");
            }
          }

          GameObject newHookGameObject = Instantiate(hookTemplateGameObject, playerRightHandTransform);
          newHookGameObject.SetActive(false);
          newHookGameObject.name = itemBase.UniqueName;
          newHookGameObject.transform.localPosition = Vector3.zero;

          Hook hookScript = newHookGameObject.GetComponent<Hook>();
          hookScript.gatherTime = AugmentedEquipmentItems.AllAugmentedEquipmentItems[itemBase.UniqueName].modifiers["gatherTime"];

          // Update the Hook's Mesh Renderer and the Mesh Filter
          foreach (Transform hookTransform in newHookGameObject.transform) {
            foreach (Transform hookModelTransform in hookTransform) {
              if (hookModelTransform.gameObject.name.Contains("Model")) {
                GameObject hookModelObject = hookModelTransform.gameObject;
                Mesh importedHookMesh = asset.LoadAsset<Mesh>(itemBase.UniqueName);
                Material importedHookMaterial = asset.LoadAsset<Material>(itemBase.UniqueName + "_Material");
                MeshFilter hookMeshFilter = hookModelObject.GetComponent<MeshFilter>();
                MeshRenderer hookMeshRenderer = hookModelObject.GetComponent<MeshRenderer>();
                hookMeshFilter.mesh = importedHookMesh;
                hookMeshRenderer.material = importedHookMaterial;
              }
            }
          }

          // Move this out to a helper method probably?
          ItemConnection itemConnection = new ItemConnection();
          itemConnection.name = itemBase.UniqueName;
          itemConnection.inventoryItem = itemBase;
          itemConnection.obj = newHookGameObject;
          connectionDictionary.TryAdd(itemBase.UniqueName, itemConnection);
          addedConnectionDictionary.Add(itemBase.UniqueName, itemConnection);

          hookThrowAnimation.objectReciever = newHookGameObject;
          animationConnections.Add(hookThrowAnimation);
          addedAnimationConnectionList.Add(hookThrowAnimation);

          hookGatherAnimation.objectReciever = newHookGameObject;
          animationConnections.Add(hookGatherAnimation);
          addedAnimationConnectionList.Add(hookGatherAnimation);
        }

        if (itemBase.UniqueName.Contains("_Shovel")) {
          GameObject shovelTemplateGameObject = null;
          foreach (Transform rightHandItem in playerRightHandTransform) {
            Shovel templateShovelScript = rightHandItem.gameObject.GetComponent<Shovel>();
            if (templateShovelScript != null) {
              shovelTemplateGameObject = rightHandItem.gameObject;
              break;
            }
          }

          AnimationConnection digDownAnimation = null;
          AnimationConnection digThrowAnimation = null;
          foreach (AnimationConnection connection in animationConnections) {
            if (connection.name.Equals("Shovel_Use | DigDown") && digDownAnimation == null) {
              digDownAnimation = new AnimationConnection();
              CopyAnimationConnection(digDownAnimation, connection);
            }
            if (connection.name.Equals("Shovel_Use | DigThrow") && digThrowAnimation == null) {
              digThrowAnimation = new AnimationConnection();
              CopyAnimationConnection(digThrowAnimation, connection);
            }
          }

          GameObject newShovelGameObject = Instantiate(shovelTemplateGameObject, playerRightHandTransform);
          newShovelGameObject.SetActive(false);
          newShovelGameObject.name = itemBase.UniqueName;
          newShovelGameObject.transform.localPosition = Vector3.zero;

          foreach (Transform shovelModelTransform in newShovelGameObject.transform) {
            if (shovelModelTransform.gameObject.name.Contains("Model")) {
              GameObject shovelModelObject = shovelModelTransform.gameObject;
              Material importedShovelMaterial = asset.LoadAsset<Material>(itemBase.UniqueName + "_Material");
              MeshRenderer shovelMeshRenderer = shovelModelObject.GetComponent<MeshRenderer>();
              shovelMeshRenderer.material = importedShovelMaterial;
            }
          }

          // Move this out to a helper method probably?
          ItemConnection itemConnection = new ItemConnection();
          itemConnection.name = itemBase.UniqueName;
          itemConnection.inventoryItem = itemBase;
          itemConnection.obj = newShovelGameObject;
          connectionDictionary.TryAdd(itemBase.UniqueName, itemConnection);
          addedConnectionDictionary.Add(itemBase.UniqueName, itemConnection);

          digDownAnimation.objectReciever = newShovelGameObject;
          animationConnections.Add(digDownAnimation);
          addedAnimationConnectionList.Add(digDownAnimation);

          digThrowAnimation.objectReciever = newShovelGameObject;
          animationConnections.Add(digThrowAnimation);
          addedAnimationConnectionList.Add(digThrowAnimation);
        }
      }

      // Invoke the SetIndexLocalPlayer code based on the original conditions of the Raft code.
      if (Semih_Network.IsHost && (GameManager.IsInNewGame || !__instance.IsLocalPlayer)) {
        __instance.SetIndexLocalPlayer();
      }

      networkPlayer.GetComponentInChildren<PlayerFOVManager>().Initialize();
      RAPI.GetLocalPlayer().GetComponentInChildren<PlayerFOVManager>().Initialize();
    }
  }

  [HarmonyPatch(typeof(Player), "SceneEventInterface.OnSceneEvent", new[] { typeof(SceneEvent) })]
  class PlayerOnSceneEvent_Patch {
    private static void Prefix(SceneEvent sceneEvent) {
      if (sceneEvent == SceneEvent.LeaveGame) {
        Network_Player networkPlayer = RAPI.GetLocalPlayer();
        if (networkPlayer.IsLocalPlayer) {
          UnloadAllAddedConnections(networkPlayer);
        }
      }
    }
  }

  [HarmonyPatch(typeof(Hotbar), "SelectHotslot", new[] { typeof(Slot) })]
  class HotbarSelectHotslot_Patch {
    private static void Postfix(Hotbar __instance, Slot slot) {
      if (!__instance.playerNetwork.IsLocalPlayer) {
        return;
      }
      if (slot == null || slot.IsEmpty) {
        return;
      }
      if (cachedLocalAnimator == null) {
        PlayerAnimator playerAnimator = __instance.playerNetwork.Animator;
        cachedLocalAnimator = playerAnimator.anim;
      }

      ItemInstance itemInstance = slot.itemInstance;
      if (itemInstance.UniqueName.Contains("_Shovel")) { 
        cachedLocalAnimator.speed = 1f + (1f * AugmentedEquipmentItems.Get(itemInstance.UniqueName).modifiers["digSpeedModifier"]);
      } else {
        cachedLocalAnimator.speed = 1f;
      }
    }
  }
}
